

class DictCache:

    def __init__(self, cache_size):
        self._cache_size = cache_size
        self._key_order_list = []
        self._cache_dict = {}

    def __getitem__(self, key):
        if self.is_cached(key):
            return self._cache_dict[key]
        else:
            raise KeyError(f"The specified key {key} is no cached")

    def is_cached(self, key) -> bool:
        return key in self._cache_dict

    def cache(self, key, value):
        if self.is_cached(key):
            self.uncache(key)

        self._key_order_list.append(key)
        self._cache_dict[key] = value

        self._handle_size()

    def uncache(self, key):
        if key in self._key_order_list:
            self._key_order_list.remove(key)
            self._cache_dict.pop(key)

    def clear_cache(self):
        while len(self._cache_dict) > 0:
            first_key = self._key_order_list[0]
            self.uncache(first_key)

    def _handle_size(self):
        while len(self._cache_dict) > self._cache_size:
            first_key = self._key_order_list[0]
            self.uncache(first_key)



import cv2 as cv
import numpy as np


def generate_test_video(frames=1000):
	"""
	Generador de imágenes donde cada imagen tiene el índice de su posición dibujado en el centro
	y codificado en el pixel superior izquierdo.
	Para decodificar el índice se debe usar lo siguiente:
		pixel = image[0][0]
		indice = pixel[0]*256**2 + pixel[1]*256 + pixel[2]
	:param frames: número de frames que se deseen generar.
	:return:
	"""
	shape = (64, 96, 3)
	for i in range(frames):
		img = np.zeros(shape)
		cv.putText(img, str(i), (32, 48), cv.FONT_HERSHEY_SIMPLEX, 1, (255, 255, 255), 4)
		# codificar en el pixel superior izquierdo el índice del frame
		img[0, 0] = (i // 256**2, (i % 256**2) // 256, i % 256)
		yield img.astype(np.uint8)


def parse_test_video_pixel(pixel):
	return pixel[0]*256**2 + pixel[1]*256 + pixel[2]


def write_video(output_dir, images, img_size, fps=20):
	out = cv.VideoWriter(output_dir, -1, fps, img_size)
	for img in images:
		out.write(img)
	out.release()


def get_format(video_path: str) -> tuple:
	cap = cv.VideoCapture(video_path)
	fourcc = int(cap.get(cv.CAP_PROP_FOURCC))
	c1, c2, c3, c4 = fourcc & 0xFF, fourcc & 0xFF00 >> 8, fourcc & 0xFF0000 >> 16, fourcc & 0xFF000000 >> 24
	cap.release()
	return c1, c2, c3, c4



import cv2 as cv
from datetime import datetime
import logging
import math
import numpy as np
import os
from os import path
import pandas as pd
import random
import requests
from timeit import default_timer as timer
import tkinter


def timeit(method):
	"""
	Decorador para medir el tiempo de ejecución de funciones
	:param method:
	:return:
	"""
	def timed(*args, **kw):
		ts = timer()
		result = method(*args, **kw)
		te = timer()
		logging.info('Time for method \'{}\': {} ms'.format(method.__name__, te - ts))
		return result
	return timed


def warmup(method):
	"""
	Decorador para realizar fase de calentamiento antes de cualquier prueba de tiempo.
	:param method:
	:return:
	"""
	def warm(*args, **kw):
		print("warming...")
		for _ in range(10000):
			random.randint(0, 1000000)
		result = method(*args, **kw)
		return result
	return warm


def load_scores(file_path):
	"""
	Cargar scores guardados en un fichero de texto, donde cada linea es un score.
	:param file_path: dirección del fichero de texto.
	:return:
	"""
	f = open(file_path, "r")
	for line in f:
		dist = float(line.strip())
		yield dist
	f.close()


def date_string():
	"""
	Obtener string con la fecha actual en formato dia-mes-año_hora-minuto-segundo.
	:return: string con la fecha.
	"""
	return datetime.now().strftime("%y-%m-%d_%H-%M-%S")


def decompose_seconds(seconds):
	"""
	Descomponer una cantidad de segundos en horas, minutos y segundos.
	:param seconds:
	:return:
	"""
	h = int(seconds // 3600)
	m = int(seconds % 3600 // 60)
	s = int(seconds % 60)
	ms = int((seconds - int(seconds)) * 1000)
	return h, m, s, ms


def time_string(seconds, with_ms=False, sep=':'):
	"""
	Devolver un string en formato HH:MM:SS según la cantidad de segundos pasados por parámetro.
	Si with_ms=True el formato es  HH:MM:SS,MS
	:param seconds:
	:param with_ms:
	:param sep:
	:return:
	"""
	h, m, s, ms = decompose_seconds(seconds)
	if with_ms:
		return f'{h:02d}{sep}{m:02d}{sep}{s:02d},{ms:03d}'
	else:
		return f'{h:02d}{sep}{m:02d}{sep}{s:02d}'


def write_xlsx(datas, output_file):
	"""
	Escribir una tabla de datos en un archivo xlsx.
	:param datas: lista de tablas de datos con formato {columna: [valores], ...}.
	:param output_file: fichero de salida.
	:return:
	"""
	file_dir, _ = path.split(output_file)
	if file_dir != '' and not path.exists(file_dir):
		os.makedirs(file_dir)
	writer = pd.ExcelWriter(output_file, engine='xlsxwriter')
	sheet_name = 'Sheet1'
	row = 0
	if isinstance(datas, list):
		for data in datas:
			df = pd.DataFrame(data)
			df.to_excel(writer, sheet_name=sheet_name, startrow=row, index=False)
			row += df.shape[0] + 2
	elif isinstance(datas, dict):
		for name, data in datas.items():
			df = pd.DataFrame(data)
			df.to_excel(writer, sheet_name=sheet_name, startrow=(row + 1), index=False)
			sheet = writer.sheets[sheet_name]
			sheet.write(row, 0, name)
			row += df.shape[0] + 3
	writer.save()


def sort_insert(sorted_list, element, key=None, descendent=False):
	"""
	Insertar ordenadamente un elemento en una lista ordenada.
	:param sorted_list: lista ordenada.
	:param element: elemento a insertar.
	:param key: criterio de orden.
	:param descendent: orden de inserción. False->ascendente, True->descendente
	:return:
	"""
	inserted = False
	for i in range(0, len(sorted_list)):
		v = sorted_list[i]
		cv = v
		ce = element
		if key:
			ce = key(element)
			cv = key(v)
		if cv < ce if descendent else cv > ce:
			sorted_list.insert(i, element)
			inserted = True
			break
	if not inserted:  # si no se insertó
		sorted_list.append(element)
	return sorted_list


def sort_insert_binary(ordered_list, element, key=None, descending=False):
	length = len(ordered_list)
	if length == 0:
		ordered_list.append(element)
	else:
		insertado = False
		inf = 0
		sup = length - 1

		while not insertado and inf <= sup:
			med = (inf + sup) // 2
			m = ordered_list[med]
			cm = m if not key else key(m)
			ce = element if not key else key(element)

			if ce == cm:
				ordered_list.insert(med + 1, element)
				insertado = True
			elif cm > ce if descending else cm < ce:
				inf = med + 1
				if inf > sup:
					ordered_list.insert(inf, element)
			else:
				sup = med - 1
				if inf > sup:
					ordered_list.insert(med, element)


def search_binary(ordered_list, element, key=None, descending=False):
	length = len(ordered_list)
	if length == 0:
		return None
	else:
		inf = 0
		sup = length - 1

		while True:
			med = (inf + sup) // 2
			m = ordered_list[med]
			cm = m if not key else key(m)
			ce = element if not key else key(element)

			if ce == cm:
				return m
			elif cm > ce if descending else cm < ce:
				inf = med + 1
				if inf > sup:
					return None
			else:
				sup = med - 1
				if inf > sup:
					return None


def normalize_values(values, range_min=None, range_max=None, target_min=0, target_max=1):
	mn = min(values) if range_min is None else range_min
	mx = max(values) if range_max is None else range_max
	dif = mx - mn
	target_dif = target_max - target_min
	return [((x - mn) / dif) * target_dif + target_min for x in values]


def screen_resolution():
	root = tkinter.Tk()
	width = root.winfo_screenwidth()
	height = root.winfo_screenheight()
	return width, height


def stack_images(images, cols=3, resize=True):
	sw, sh = screen_resolution()  # resolución de la pantalla
	n = len(images)  # cantidad de imágenes
	cols = min(n, cols)  # limitar el número de columnas en caso de que sean menos imágenes
	rows = math.ceil(n / cols)  # cantidad de filas
	if resize:
		nsize = (int((sw - 50) / cols), int((sh - 80) / rows))  # nuevo tamaño de cada imagen para que quepan en pantalla
		images = [cv.resize(img, nsize) for img in images]  # cambiar el tamaño de todas las imágenes

	# rellenar con imágenes en negro
	total_imgs = cols * rows
	count_fill = total_imgs - n  # cantidad a rellenar
	if count_fill > 0:
		shape = images[0].shape
		dtype = images[0].dtype
		black_img = np.zeros(shape, dtype)
		for _ in range(count_fill):
			images.append(black_img)

	new_shape = np.concatenate(((rows, cols), images[0].shape))
	images = np.array(images)
	reshaped = images.reshape(new_shape)

	vstacks = []
	for row_imgs in reshaped:
		row_stack = np.hstack(row_imgs)
		vstacks.append(row_stack)

	stack = np.vstack(vstacks)
	return stack


def stack_and_write(imgs_path, out):
	images = [cv.imread(img) for img in imgs_path]

	stack = stack_images(images)
	cv.imwrite(out, stack)


def take_skip(iterable, take, skip):
	"""
	Método para modificar la manera de iterar sobre una estructura de datos iterable, de manera que se tomen 'take' elementos
	y se ignoren 'skip' elementos de manera sucesiva. Ejemplo: si se tiene la lista l=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10] y se
	invoca take_skip(l, 3, 4), la salida sería un generador que devuelve los elementos [1, 2, 3, 8, 9, 10]
	:param self:
	:param iterable:
	:param take:
	:param skip:
	:return:
	"""
	assert take > 0, 'El parámetro take tiene que ser mayor que 0.'
	temp_take = take
	temp_skip = skip
	for x in iterable:
		if temp_take > 0 or skip == 0:
			yield x
			temp_take -= 1
		else:
			if temp_skip > 1:
				temp_skip -= 1
			else:
				temp_take = take
				temp_skip = skip


def mk_string(collection, sep=', ', start='', end=''):
	ret = start
	for item in collection:
		ret = ret + str(item) + sep
	# Removing the last inserted separator.
	ret = ret[:len(ret) - len(sep)]
	return ret + end


def shutdown():
	os.system('shutdown /p /f')


def download_file(url, output):
	r = requests.get(url)
	with open(output, 'wb') as f:
		f.write(r.content)


def download_files(dict_url_output):
	with requests.Session() as s:
		i = 0
		for url, output in dict_url_output.items():
			r = s.get(url)
			with open(output, 'wb') as f:
				f.write(r.content)
			i += 1
			if i % 250 == 0: print('downloaded {} images'.format(i))



import numpy as np
import cv2 as cv


def validate_box(box, img_width, img_height):
	"""
	Validar que las dimensiones de un rectángulo no estén fuera de las dimensiones de una imagen.
	:param x1: límite izquierdo del rectángulo.
	:param y1: límite superior del rectángulo.
	:param x2: límite derecho del rectángulo.
	:param y2: límite inferior del rectángulo.
	:param img_width: ancho de la imagen.
	:param img_height: alto de la imagen.
	:return: True si el rectángulo es válido, False en caso contrario.
	"""
	x1, y1, x2, y2 = box
	if x1 < 0 or y1 < 0 or x2 < 0 or y2 < 0:
		return False
	if x1 > img_width or y1 > img_height or x2 > img_width or y2 > img_height:
		return False
	return True


def validate_boxes(boxes, img_width, img_height):
	"""
	De una lista de rectángulos, filtra los que son válidos, es decir, los que sus dimensiones no se salen de las dimensiones
	de la imagen.
	:param boxes: lista de rectángulos.
	:param img_width: ancho de la imagen.
	:param img_height: alto de la imagen.
	:return: lista filtrada.
	"""
	return [box for box in boxes if validate_box(box, img_width, img_height)]


def get_boxes(detections):
	return [box for box, _ in detections]


def get_confidences(detections):
	return [confidence for _, confidence in detections]


def filter_confidence(detections, threshold):
	return [(box, confidence) for box, confidence in detections if confidence >= threshold]


def filter_valid_dets(detections, img_width, img_height):
	return [(box, confidence) for box, confidence in detections if validate_box(box, img_width, img_height)]


def filter_valid_boxes(boxes, img_width, img_height):
	"""
	De una lista de rectángulos, filtra los que son válidos, es decir, los que sus dimensiones no se salen de las dimensiones
	de la imagen.
	:param boxes: lista de rectángulos.
	:param img_width: ancho de la imagen.
	:param img_height: alto de la imagen.
	:return: lista filtrada.
	"""
	return [box for box in boxes if validate_box(box, img_width, img_height)]


def intersection_over_union(A, B):
	(x1a, y1a, x2a, y2a) = A
	(x1b, y1b, x2b, y2b) = B
	# determine the (x, y)-coordinates of the intersection rectangle
	x1I = max(x1a, x1b)
	y1I = max(y1a, y1b)
	x2I = min(x2a, x2b)
	y2I = min(y2a, y2b)

	# compute the area of intersection rectangle
	interArea = max(0, x2I - x1I) * max(0, y2I - y1I)

	# compute the area of both the prediction and ground-truth rectangles
	boxAArea = (x2a - x1a) * (y2a - y1a)
	boxBArea = (x2b - x1b) * (y2b - y1b)

	iou = interArea / float(boxAArea + boxBArea - interArea)
	return iou


def non_max_suppression(boxes, overlap_thresh=0.3, iou_thresh=0.65):
	if len(boxes) == 0:
		return []
	if boxes.dtype.kind == "i":
		boxes = boxes.astype("float")
	pick = []
	y2 = boxes[:, 3]
	idxs = np.argsort(y2)
	while len(idxs) > 0:
		last = len(idxs) - 1
		i = idxs[last]
		pick.append(i)

		iou_overlap = np.array([intersection_over_union(boxes[i], b) for b in boxes[idxs[:last]]])
		idxs = np.delete(idxs, np.concatenate(([last], np.where(iou_overlap > iou_thresh)[0])))
	return pick


def box_area(box):
	x1, y1, x2, y2 = box
	return (x2 - x1) * (y2 - y1)


def box_center(box):
	x1, y1, x2, y2 = box
	return (x2 - x1) // 2, (y2 - y1) // 2


def fabiolas_adjust_box(img, face):
	"""
	Función de Fabiola para ajustar rectángulos de rostros.
	:param img:
	:param face:
	:return:
	"""
	image_rows, image_cols = img.shape[:2]
	x1, y1, x2, y2 = face
	face_width = x2 - x1
	face_height = y2 - y1
	face_center = (x1 + face_width / 2, y1 + face_height / 2)
	dif_w = dif_h = (face_height - face_width) / 2

	half_w = (face_width + dif_w) / 2
	if face_center[0] < half_w or face_center[0] + half_w > image_cols:
		half_w = min(face_center[0], image_cols - face_center[0])
	new_x = face_center[0] - half_w

	half_h = (face_height - dif_h) / 2
	new_y = face_center[1] - half_h

	return int(new_x), int(new_y), int(new_x + 2 * half_w), int(new_y + 2 * half_h)


def crop(img, box):
	x1, y1, x2, y2 = box
	return img[y1:y2, x1:x2]


def crop_with_margin(img, face, margin=0.4):
	"""
	Recortar imagen dando un margen extra.
	:author Fabiola
	:param img: image to crop
	:param face: face rect
	:param margin: margin around face to crop image
	:return: tuple(output cropped img, coordinates of cropped rect in img, coordinates of cropped rect in output)
	"""
	img_h, img_w, d = img.shape

	x1, y1, x2, y2 = face
	w = x2 - x1
	h = y2 - y1
	x_margin = int(w * margin)
	y_margin = int(h * margin)

	# --add margin--
	ini_x = x1 - x_margin
	end_x = x2 + x_margin
	ini_y = y1 - y_margin
	end_y = y2 + y_margin

	output = np.zeros((end_y - ini_y, end_x - ini_x, d), dtype=img.dtype)

	# --ensure that margin doesn't go beyond image size--
	x = max(ini_x, 0)
	brx = min(end_x, img_w)
	y = max(ini_y, 0)
	bry = min(end_y, img_h)

	# --coordinates of the cropped rect region in img--
	crop = x, y, brx, bry

	# --coordinates of the cropped rect region in output image--
	new_x = x - ini_x
	new_brx = new_x + brx - x
	new_y = y - ini_y
	new_bry = new_y + bry - y
	padding_crop = new_x, new_y, new_brx, new_bry

	# --cropping--
	output[new_y:new_bry, new_x:new_brx, :] = img[y:bry, x:brx, :]

	# --if margin goes beyond the size of the image, repeat last row/column of pixels--
	if new_y > 0:
		cv.repeat(output[new_y:new_y + 1, 0:output.shape[1]], new_y, 1, output[0:new_y, 0:output.shape[1]])

	if new_bry < output.shape[0]:
		cv.repeat(output[new_bry - 1: new_bry, 0:output.shape[1]], output.shape[0] - new_bry, 1, output[new_bry:output.shape[0], 0:output.shape[1]])

	if new_x > 0:
		cv.repeat(output[0:output.shape[0], new_x:new_x + 1], 1, new_x, output[0:output.shape[0], 0:new_x])

	if new_brx < output.shape[1]:
		cv.repeat(output[0:output.shape[0], new_brx - 1: new_brx], 1, output.shape[1] - new_brx, output[0:output.shape[0], new_brx:output.shape[1]])

	return output, crop, padding_crop



import cv2 as cv
import numpy as np


def fit_in_size(img, size):
	"""
	Cambiar el tamaño de la imagen rellenando con color negro para mantener la relación de aspecto.
	:param img:
	:param size:
	:return:
	"""
	h, w, d = img.shape
	width, height = size
	if height <= 4 or width <= 4:
		return img

	image_aspect = w / h
	size_aspect = width / height

	resized_img = np.zeros((height, width, d), dtype=img.dtype)
	if image_aspect > size_aspect:  # top/bottom stripes
		new_h = round(width / image_aspect)
		img_copy = cv.resize(img, (width, new_h), interpolation=cv.INTER_AREA)
		offset = height - new_h
		half_offset = offset // 2
		resized_img[half_offset: half_offset + new_h, :, :] = img_copy[:, :, :]
	else:  # left/right stripes
		new_w = round(height * image_aspect)
		img_copy = cv.resize(img, (new_w, height), interpolation=cv.INTER_AREA)
		offset = width - new_w
		half_offset = offset // 2
		resized_img[:, half_offset: half_offset + new_w, :] = img_copy
	return resized_img


def empty_gray_image():
	return np.zeros((8, 8, 3), dtype=np.uint8) + 127



import os
from os import path


def create_dir_of_file(file):
	"""
	Crea el directorio de un archivo, si este no existe.
	:param file: fichero.
	:return:
	"""
	directory, _ = path.split(file)
	if not path.exists(directory):
		os.makedirs(directory)


def replace_in_all_files(old, new, extensions=None, root='datasets/'):
	files = list_files(root, validExts=extensions)
	for f in files:
		_, name = os.path.split(f)
		if old in name:
			nf = f.replace(old, new)
			os.rename(src=f, dst=nf)


image_types = (".jpg", ".jpeg", ".png", ".bmp", ".tif", ".tiff")


def list_images(basePath, contains=None, recursive=True):
	# return the set of files that are valid
	return list_files(basePath, validExts=image_types, contains=contains, recursive=recursive)


def list_files(basePath, validExts=None, contains=None, recursive=True):
	# loop over the directory structure
	for (rootDir, dirNames, filenames) in os.walk(basePath):
		# loop over the filenames in the current directory
		for filename in filenames:
			# if the contains string is not none and the filename does not contain
			# the supplied string, then ignore the file
			if contains is not None and filename.find(contains) == -1:
				continue

			# determine the file extension of the current file
			ext = filename[filename.rfind("."):].lower()

			# check to see if the file is an image and should be processed
			if validExts is None or ext.endswith(validExts):
				# construct the path to the image and yield it
				imagePath = os.path.join(rootDir, filename)
				yield imagePath
		if not recursive:
			break



import ast
import inspect


def number_callable_params(callable):
	"""
	Número de parámetros de una función o callable.
	:param callable: objeto que puede ser invocado
	:return:
	"""
	return len(inspect.signature(callable).parameters)


def function_has_return(callable):
	"""
	Devuelve True si la función tiene sentencia return, False en caso contrario.
	No funciona con funciones cuyo código no esté escrito en Python.
	:param callable:
	:return:
	"""
	return any(isinstance(node, ast.Return) for node in ast.walk(ast.parse(inspect.getsource(callable))))


def next_class_in_mro(instance_class: type, current_class: type) -> type:
	mro = instance_class.__mro__
	for i, clazz in enumerate(mro):
		if clazz == current_class:
			if i < len(mro) - 1:
				return mro[i + 1]
	return None


def exist_method_in_parent_classes(instance: object, current_class: type, method_name: str) -> bool:
	instance_class = instance.__class__
	mro = instance_class.__mro__
	for i, clazz in enumerate(mro):
		if clazz == current_class:
			next_classes = mro[i + 1:]
			for next_class in next_classes:
				if method_name in dir(next_class):
					return True
	return False



import cv2 as cv


def draw_rects(img, color, boxes, scores=None):
	"""
	Dibuja un conjunto de rectángulos en la imagen con el color especificado. Opcionalmente los rectángulos pueden tener
	una puntuación.
	:param img: immagen.
	:param boxes: rectángulos.
	:param color: color.
	:param scores: puntuaciones [opcional].
	:return:
	"""
	draw = img.copy()
	for i in range(len(boxes)):
		box = boxes[i]
		cv.rectangle(draw, (box[0], box[1]), (box[2], box[3]), color, 2, 4)
		if scores:
			text = "{:.2f}".format(scores[i])
			cv.putText(draw, text, (box[0], box[1] - 5), cv.FONT_HERSHEY_SIMPLEX, 1.5, color, 2, cv.LINE_AA)
	return draw


def draw_face_metadata(img, color, box, age, gender, ethnicity):
	draw = img.copy()
	age = age or '-'
	gender = gender or '-'
	ethnicity = ethnicity or '-'
	data_str = str.join(' | ', (age, gender, ethnicity))
	cv.rectangle(draw, (box[0], box[1]), (box[2], box[3]), color, 2, 4)
	cv.putText(draw, data_str, (box[0], box[1] - 5), cv.FONT_HERSHEY_SIMPLEX, 0.8, color, 2, cv.LINE_AA)
	return draw



from collections import Generator


class Stream:
	"""
	Estructura de datos para represantar un stream de datos que mantiene en memoria los elementos ya visitados.
	"""

	def __init__(self, generator: Generator):
		"""
		Constructor del stream. Recibe un generador que representa la fuente del stream
		:param generator:
		"""
		self._generator = generator
		self._data = []
		self.__last_index = 0
		self.__generator_ended = False

	def __getitem__(self, item):
		stop = item
		if isinstance(item, slice):
			stop = item.stop or -1
			if item.start is not None and item.start < 0:
				stop = -1

		if stop < 0:
			while not self.__generator_ended:
				self.__append_next_value()
		else:
			for i in range(self.__last_index, stop + 1):
				self.__append_next_value()

		return self._data[item]

	def __append_next_value(self):
		if not self.__generator_ended:
			try:
				nv = next(self._generator)
				self._data.append(nv)
				self.__last_index += 1
			except StopIteration:
				self.__generator_ended = True


if __name__ == '__main__':
	s = Stream(None)
	for x in s:
		print(x)
